"use strict";
// Copyright (c) Microsoft Corporation.
// Licensed under the MIT License.
Object.defineProperty(exports, "__esModule", { value: true });
exports.checkForNameCollisions = exports.generateClient = void 0;
const operationGenerator_1 = require("./operationGenerator");
const nameUtils_1 = require("../utils/nameUtils");
const codemodel_1 = require("@autorest/codemodel");
const parameterUtils_1 = require("./utils/parameterUtils");
const importUtils_1 = require("./utils/importUtils");
const responseTypeUtils_1 = require("./utils/responseTypeUtils");
const tracingUtils_1 = require("./utils/tracingUtils");
const pagingOperations_1 = require("./utils/pagingOperations");
const autorestSession_1 = require("../autorestSession");
function generateClient(clientDetails, project) {
    const { useCoreV2, hideClients, srcPath, addCredentials, packageDetails } = autorestSession_1.getAutorestOptions();
    const hasMappers = !!clientDetails.mappers.length;
    // Check if there are any client level operations
    const inlineOperations = clientDetails.operationGroups.filter(og => og.isTopLevel);
    const hasInlineOperations = inlineOperations.length > 0;
    // Check if there are any non client-level operations to import
    const importedOperations = clientDetails.operationGroups.filter(og => !og.isTopLevel);
    const hasImportedOperations = importedOperations.length > 0;
    if (hasImportedOperations && hasInlineOperations) {
        // Check if there is a name collision between client-level operation names
        // and operation group key names.
        checkForNameCollisions(importedOperations, inlineOperations);
    }
    const hasCredentials = !!addCredentials;
    const hasClientOptionalParams = clientDetails.parameters.some(p => !p.required && p.implementationLocation === codemodel_1.ImplementationLocation.Client);
    const clientFile = project.createSourceFile(`${srcPath}/${clientDetails.sourceFileName}.ts`, undefined, {
        overwrite: true
    });
    !useCoreV2 && writePackageInfo(clientFile, packageDetails);
    const flattenedInlineOperations = inlineOperations.reduce((acc, curr) => (acc = [...acc, ...curr.operations]), []);
    if (!useCoreV2) {
        clientFile.addImportDeclaration({
            namespaceImport: "coreHttp",
            moduleSpecifier: "@azure/core-http"
        });
    }
    else {
        clientFile.addImportDeclaration({
            namespaceImport: "coreClient",
            moduleSpecifier: "@azure/core-client"
        });
    }
    if (hasCredentials || hasInlineOperations || !hasClientOptionalParams) {
        clientFile.addImportDeclaration({
            namespaceImport: "coreRestPipeline",
            moduleSpecifier: "@azure/core-rest-pipeline"
        });
        clientFile.addImportDeclaration({
            namespaceImport: "coreTracing",
            moduleSpecifier: "@azure/core-tracing"
        });
        if (hasCredentials) {
            clientFile.addImportDeclaration({
                namespaceImport: "coreAuth",
                moduleSpecifier: "@azure/core-auth"
            });
        }
    }
    pagingOperations_1.addPagingEsNextRef(flattenedInlineOperations, clientFile);
    pagingOperations_1.addPagingImports(flattenedInlineOperations, clientFile);
    const hasLro = inlineOperations.some(og => og.operations.some(o => o.isLro));
    if (hasInlineOperations && hasLro) {
        clientFile.addImportDeclaration({
            namedImports: ["PollerLike", "PollOperationState", "LroEngine"],
            moduleSpecifier: "@azure/core-lro"
        });
        clientFile.addImportDeclaration({
            namedImports: ["LroImpl"],
            moduleSpecifier: `./lroImpl`
        });
    }
    if (hasImportedOperations) {
        clientFile.addImportDeclaration({
            namedImports: importedOperations.map(o => `${nameUtils_1.normalizeName(o.name, nameUtils_1.NameType.OperationGroup, true /* shouldGuard */)}Impl`),
            moduleSpecifier: "./operations"
        });
        clientFile.addImportDeclaration({
            namedImports: importedOperations.map(o => `${nameUtils_1.normalizeName(o.name, nameUtils_1.NameType.OperationGroup, true /* shouldGuard */)}`),
            moduleSpecifier: "./operationsInterfaces"
        });
    }
    if (hasInlineOperations && importUtils_1.shouldImportParameters(clientDetails)) {
        tracingUtils_1.addTracingOperationImports(clientFile, ".");
        clientFile.addImportDeclaration({
            namespaceImport: "Parameters",
            moduleSpecifier: "./models/parameters"
        });
    }
    // Only import mappers if there are any
    if (hasInlineOperations && hasMappers) {
        clientFile.addImportDeclaration({
            namespaceImport: "Mappers",
            moduleSpecifier: "./models/mappers"
        });
    }
    const clientClass = clientFile.addClass({
        name: clientDetails.className,
        extends: !useCoreV2 ? "coreHttp.ServiceClient" : "coreClient.ServiceClient",
        isExported: true
    });
    if (hideClients) {
        clientClass.addJsDoc({
            tags: [
                {
                    tagName: "internal"
                }
            ]
        });
    }
    const importedModels = new Set();
    const clientParams = clientDetails.parameters.filter(param => param.implementationLocation === codemodel_1.ImplementationLocation.Client);
    writeClassProperties(clientClass, clientParams, importedModels);
    writeConstructor(clientDetails, clientClass, importedModels);
    writeClientOperations(clientFile, clientClass, clientDetails, hasLro, importedModels);
    // Use named import from Models
    if (importedModels.size) {
        clientFile.addImportDeclaration({
            namedImports: [...importedModels],
            moduleSpecifier: "./models"
        });
    }
    clientFile.fixUnusedIdentifiers();
}
exports.generateClient = generateClient;
function writeClassProperties(clientClass, clientParams, importedModels) {
    const params = clientParams.filter(p => !p.isSynthetic);
    params.forEach(({ typeDetails }) => typeDetails.usedModels.forEach(model => importedModels.add(model)));
    clientClass.addProperties(params.map(param => {
        return {
            name: param.name,
            type: param.typeDetails.typeName,
            hasQuestionToken: !param.required
        };
    }));
}
function checkForNameCollisions(importedOperations, inlineOperations) {
    const groupOpsKeyNames = importedOperations.map(og => og.key.toLowerCase());
    const inlineOpsKeyNames = inlineOperations.map(og => og.operations.map(operation => operation.name.toLowerCase()));
    const collidingKeyNames = inlineOpsKeyNames.map(inlineOpsKeyArray => inlineOpsKeyArray.filter(inlineOpKey => groupOpsKeyNames.includes(inlineOpKey)));
    if (collidingKeyNames.length > 0 && collidingKeyNames[0].length > 0) {
        const messages = collidingKeyNames.map(key => `Operation Group(s) '${key}' is/are colliding with client-level operation(s) with the same name.`);
        throw new Error(...messages);
    }
}
exports.checkForNameCollisions = checkForNameCollisions;
function writeConstructor(clientDetails, classDeclaration, importedModels) {
    const requiredParams = clientDetails.parameters.filter(param => param.required &&
        param.implementationLocation === codemodel_1.ImplementationLocation.Client &&
        !param.defaultValue &&
        param.schemaType !== codemodel_1.SchemaType.Constant);
    const hasClientOptionalParameters = clientDetails.parameters.some(param => !param.required &&
        param.implementationLocation === codemodel_1.ImplementationLocation.Client);
    const docs = [
        `Initializes a new instance of the ${clientDetails.className} class.`,
        ...parameterUtils_1.formatJsDocParam(requiredParams),
        `@param options The parameter options`
    ]
        .filter(d => !!d)
        .join("\n");
    let optionsParameterType = "ServiceClientOptions";
    if (hasClientOptionalParameters) {
        const paramType = `${clientDetails.className}OptionalParams`;
        importedModels.add(paramType);
        optionsParameterType = paramType;
    }
    requiredParams.forEach(({ typeDetails }) => typeDetails.usedModels.forEach(model => importedModels.add(model)));
    const clientConstructor = classDeclaration.addConstructor({
        docs: [docs],
        parameters: [
            ...requiredParams.map(p => ({
                name: p.name,
                hasQuestionToken: !p.required,
                type: p.typeDetails.typeName
            })),
            {
                name: "options",
                hasQuestionToken: true,
                type: optionsParameterType
            }
        ]
    });
    const { useCoreV2 } = autorestSession_1.getAutorestOptions();
    const hasLro = clientDetails.operationGroups.some(og => og.operations.some(o => o.isLro));
    const clientParams = clientDetails.parameters.filter(param => param.implementationLocation === codemodel_1.ImplementationLocation.Client);
    const addBlankLine = true;
    const requiredParameters = getRequiredParamAssignments(requiredParams);
    const constantParameters = getConstantClientParamAssignments(clientParams);
    const writeStatement = (content, shouldAddBlankLine = false) => (writer) => {
        if (content) {
            writer.writeLine(content);
            shouldAddBlankLine && writer.blankLine();
        }
    };
    const writeStatements = (lines, shouldAddBlankLine = false) => (writer) => {
        lines.forEach(line => writer.writeLine(line));
        shouldAddBlankLine && writer.blankLine();
    };
    clientConstructor.addStatements([
        writeStatements(getRequiredParamChecks(requiredParams), addBlankLine),
        writeStatement(writeDefaultOptions(clientParams.some(p => p.name === "credentials"), hasLro, clientDetails))
    ]);
    !useCoreV2 &&
        clientConstructor.addStatements([
            writeStatement(getEndpointStatement(clientDetails.endpoint), addBlankLine)
        ]);
    clientConstructor.addStatements([
        requiredParameters.length ? "// Parameter assignments" : "",
        writeStatements(getRequiredParamAssignments(requiredParams), addBlankLine),
        constantParameters.length
            ? "// Assigning values to Constant parameters"
            : "",
        writeStatements(constantParameters, addBlankLine)
    ]);
    const operationDeclarationDetails = getOperationGroupsDeclarationDetails(clientDetails.operationGroups.filter(og => !og.isTopLevel));
    clientConstructor.addStatements([
        ...operationDeclarationDetails.map(({ name, typeName }) => `this.${name} = new ${typeName}Impl(this)`)
    ]);
}
function getOperationGroupsDeclarationDetails(operationGroups) {
    return operationGroups.map(og => {
        return {
            name: nameUtils_1.normalizeName(og.name, nameUtils_1.NameType.Property),
            typeName: `${nameUtils_1.normalizeName(og.name, nameUtils_1.NameType.OperationGroup, true /* shouldGuard */)}`
        };
    });
}
function writeClientOperations(file, classDeclaration, clientDetails, hasLro, importedModels) {
    const allModelsNames = responseTypeUtils_1.getAllModelsNames(clientDetails);
    const topLevelGroup = clientDetails.operationGroups.find(og => og.isTopLevel);
    const hasMappers = !!clientDetails.mappers.length;
    // Add top level operation groups as client properties
    if (!!topLevelGroup) {
        if (hasLro) {
            operationGenerator_1.writeGetOperationOptions(classDeclaration);
        }
        operationGenerator_1.writeOperations(topLevelGroup, classDeclaration, importedModels, allModelsNames, clientDetails, true // isInline
        );
        operationGenerator_1.addOperationSpecs(topLevelGroup, file, clientDetails.parameters, hasMappers);
    }
    const operationsDeclarationDetails = getOperationGroupsDeclarationDetails(clientDetails.operationGroups.filter(og => !og.isTopLevel));
    // Each operation group will have its class
    // and the client class will have each group as properties
    classDeclaration.addProperties(operationsDeclarationDetails.map(op => {
        return {
            name: op.name,
            type: op.typeName
        };
    }));
}
function getRequiredParamChecks(requiredParameters) {
    return requiredParameters.map(({ name }) => `if(${name} === undefined) {
    throw new Error("'${name}' cannot be null");
  }`);
}
function getCredentialScopesValue(credentialScopes) {
    if (Array.isArray(credentialScopes)) {
        return `[${credentialScopes.map(scope => `"${scope}"`).join()}]`;
    }
    else if (typeof credentialScopes === "string") {
        return `"${credentialScopes}"`;
    }
    return credentialScopes;
}
function getTrack2DefaultContent(addScopes, hasCredentials) {
    return `// Initializing default values for options
  if (!options) {
     options = {};
  }

  const defaultUserAgent = \`azsdk-js-\${packageName.replace(/@.*\\//,"")}/\${packageVersion} \${coreHttp.getDefaultUserAgentValue()}\`;
  
  ${addScopes}

  super(${hasCredentials ? "credentials" : `undefined`}, {
    ...options,
    userAgent: options.userAgent
      ? \`\${options.userAgent} \${defaultUserAgent}\`
      : \`\${defaultUserAgent}\`
  });
  
  this.requestContentType = "application/json; charset=utf-8";
  
  `;
}
function getTrack1DefaultContent(addScopes, defaults, packageDetails, clientDetails) {
    return `// Initializing default values for options
  if (!options) {
    options = {};
  }
  ${defaults}

  const packageDetails = \`azsdk-js-${packageDetails.name.replace(/@.*\//, "")}/${packageDetails.version}\`;
  const userAgentPrefix =
      options.userAgentOptions && options.userAgentOptions.userAgentPrefix
        ? \`\${options.userAgentOptions.userAgentPrefix} \${packageDetails}\`
        : \`\${packageDetails}\`;
  
  ${addScopes}
  const optionsWithDefaults = {
    ...defaults,
    ...options,
    userAgentOptions: {
      userAgentPrefix
    },
    baseUri: ${getEndpoint(clientDetails.endpoint)}
  };
  super(optionsWithDefaults);
  `;
}
function writeDefaultOptions(hasCredentials, hasLro, clientDetails) {
    const { useCoreV2, credentialScopes, packageDetails } = autorestSession_1.getAutorestOptions();
    const credentialScopesValues = getCredentialScopesValue(credentialScopes);
    const addScopes = credentialScopes
        ? `if(!options.credentialScopes) {
    options.credentialScopes = ${credentialScopesValues}
  }`
        : "";
    const defaults = !hasCredentials
        ? `const defaults: ${clientDetails.className}OptionalParams = {
    requestContentType: "application/json; charset=utf-8"
  };`
        : `const defaults: ${clientDetails.className}OptionalParams = {
    requestContentType: "application/json; charset=utf-8",
    credential: credentials
  };`;
    return !useCoreV2
        ? getTrack2DefaultContent(addScopes, hasCredentials)
        : getTrack1DefaultContent(addScopes, defaults, packageDetails, clientDetails);
}
function getEndpointStatement({ endpoint }) {
    return `this.baseUri = options.endpoint ${endpoint ? ` || "${endpoint}"` : ""};`;
}
function getEndpoint({ endpoint }) {
    return `options.endpoint ${endpoint ? ` || "${endpoint}"` : ""}`;
}
function getRequiredParamAssignments(requiredParameters) {
    const disallowedClientParameters = ["credentials"];
    return requiredParameters
        .filter(({ name }) => !disallowedClientParameters.includes(name))
        .map(({ name }) => `this.${name} = ${name};`);
}
function getConstantClientParamAssignments(clientParameters) {
    return clientParameters
        .filter(p => !!p.defaultValue || p.schemaType === codemodel_1.SchemaType.Constant)
        .map(({ name, defaultValue }) => `this.${name} = options.${name} ||  ${defaultValue}`);
}
function writePackageInfo(sourceFile, packageDetails) {
    sourceFile.addStatements([
        `\n\n`,
        `const packageName = "${packageDetails.name || ""}";`,
        `const packageVersion = "${packageDetails.version || ""}";`
    ]);
}
//# sourceMappingURL=clientFileGenerator.js.map