"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.generateIndexFile = void 0;
const autorestSession_1 = require("../autorestSession");
const nameUtils_1 = require("../utils/nameUtils");
const extractPaginationDetails_1 = require("../utils/extractPaginationDetails");
const hasPollingOperations_1 = require("../restLevelClient/helpers/hasPollingOperations");
const modelHelpers_1 = require("../restLevelClient/helpers/modelHelpers");
function generateIndexFile(project, clientDetails) {
    const { restLevelClient, srcPath, batch } = autorestSession_1.getAutorestOptions();
    const indexFile = project.createSourceFile(`${srcPath}/index.ts`, undefined, {
        overwrite: true
    });
    if (!restLevelClient) {
        if (!clientDetails) {
            throw new Error("ClientDetails are required when generating High Level Clients");
        }
        generateHLCIndex(clientDetails, indexFile);
    }
    else if (!batch || (batch === null || batch === void 0 ? void 0 : batch.length) === 1) {
        // if we are generate single client package for RLC
        generateRLCIndex(indexFile);
    }
    else {
        generateRLCIndexForMultiClient(indexFile);
    }
}
exports.generateIndexFile = generateIndexFile;
// to generate a index.ts for each single module inside the multi client RLC package
function generateRLCIndexForMultiClient(file) {
    const { model } = autorestSession_1.getSession();
    const clientName = model.language.default.name;
    const moduleName = nameUtils_1.normalizeName(clientName, nameUtils_1.NameType.File);
    file.addImportDeclaration({
        namespaceImport: "Parameters",
        moduleSpecifier: "./parameters"
    });
    file.addImportDeclaration({
        namespaceImport: "Responses",
        moduleSpecifier: "./responses"
    });
    file.addImportDeclaration({
        namespaceImport: "Client",
        moduleSpecifier: "./clientDefinitions"
    });
    const exports = ['Parameters', 'Responses', 'Client'];
    if (modelHelpers_1.hasInputModels(model)) {
        file.addImportDeclaration({
            namespaceImport: "Models",
            moduleSpecifier: "./models",
        });
        exports.push("Models");
    }
    if (modelHelpers_1.hasOutputModels(model)) {
        file.addImportDeclaration({
            namespaceImport: "OutputModels",
            moduleSpecifier: "./outputModels",
        });
        exports.push("OutputModels");
    }
    if (extractPaginationDetails_1.hasPagingOperations(model)) {
        file.addImportDeclaration({
            namespaceImport: "PaginateHelper",
            moduleSpecifier: './paginateHelper',
        });
        exports.push("PaginateHelper");
    }
    if (hasPollingOperations_1.hasPollingOperations(model)) {
        file.addImportDeclaration({
            namespaceImport: 'PollingHelper',
            moduleSpecifier: './pollingHelper',
        });
        exports.push("PollingHelper");
    }
    file.addExportDeclarations([
        {
            moduleSpecifier: `./${moduleName}`,
            namedExports: [`${clientName} as ${clientName}Client`],
        },
        {
            namedExports: [...exports]
        }
    ]);
}
function generateRLCIndex(file) {
    const { model } = autorestSession_1.getSession();
    const clientName = model.language.default.name;
    const moduleName = nameUtils_1.normalizeName(clientName, nameUtils_1.NameType.File);
    file.addImportDeclaration({
        moduleSpecifier: `./${moduleName}`,
        defaultImport: clientName
    });
    file.addExportDeclarations([
        {
            moduleSpecifier: `./${moduleName}`,
        },
        {
            moduleSpecifier: "./parameters"
        },
        {
            moduleSpecifier: "./responses"
        },
        {
            moduleSpecifier: "./clientDefinitions"
        }
    ]);
    if (modelHelpers_1.hasInputModels(model)) {
        file.addExportDeclarations([
            {
                moduleSpecifier: "./models"
            }
        ]);
    }
    if (modelHelpers_1.hasOutputModels(model)) {
        file.addExportDeclarations([
            {
                moduleSpecifier: "./outputModels"
            }
        ]);
    }
    if (extractPaginationDetails_1.hasPagingOperations(model)) {
        file.addExportDeclarations([
            {
                moduleSpecifier: "./paginateHelper"
            }
        ]);
    }
    if (hasPollingOperations_1.hasPollingOperations(model)) {
        file.addExportDeclarations([
            {
                moduleSpecifier: "./pollingHelper"
            }
        ]);
    }
    file.addExportAssignment({
        expression: clientName,
        isExportEquals: false
    });
}
function generateHLCIndex(clientDetails, file) {
    const { disablePagingAsyncIterators } = autorestSession_1.getAutorestOptions();
    if (clientDetails.options.hasPaging && !disablePagingAsyncIterators) {
        file.addStatements([`/// <reference lib="esnext.asynciterable" />`]);
    }
    file.addExportDeclarations([
        {
            moduleSpecifier: "./models"
        },
        {
            moduleSpecifier: `./${clientDetails.sourceFileName}`,
            namedExports: [clientDetails.className]
        }
    ]);
    const operationGroups = clientDetails.operationGroups.filter(og => !og.isTopLevel);
    if (operationGroups.length) {
        file.addExportDeclarations([
            {
                moduleSpecifier: "./operationsInterfaces"
            }
        ]);
    }
    file.fixUnusedIdentifiers();
}
//# sourceMappingURL=indexGenerator.js.map