"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.generateClient = void 0;
const codemodel_1 = require("@autorest/codemodel");
const ts_morph_1 = require("ts-morph");
const path = require("path");
const autorestSession_1 = require("../autorestSession");
const urlTransforms_1 = require("../transforms/urlTransforms");
const nameUtils_1 = require("../utils/nameUtils");
const schemaHelpers_1 = require("./schemaHelpers");
const languageHelpers_1 = require("../utils/languageHelpers");
const generateMethodShortcuts_1 = require("./generateMethodShortcuts");
const generateClientDefinition_1 = require("./generateClientDefinition");
function generateClient(model, project) {
    const name = nameUtils_1.normalizeName(languageHelpers_1.getLanguageMetadata(model.language).name, nameUtils_1.NameType.File);
    const { srcPath } = autorestSession_1.getAutorestOptions();
    const clientFile = project.createSourceFile(path.join(srcPath, `${name}.ts`), undefined, {
        overwrite: true
    });
    // Get all paths
    const clientName = languageHelpers_1.getLanguageMetadata(model.language).name;
    const uriParameter = getClientUriParameter();
    const { addCredentials, credentialKeyHeaderName, batch } = autorestSession_1.getAutorestOptions();
    const credentialTypes = addCredentials ? ["TokenCredential"] : [];
    if (credentialKeyHeaderName) {
        credentialTypes.push("KeyCredential");
    }
    const commonClientParams = [
        ...(uriParameter ? [{ name: uriParameter, type: "string" }] : []),
        ...(addCredentials === false
            ? []
            : [{ name: "credentials", type: credentialTypes.join(" | ") }])
    ];
    const clientInterfaceName = `${clientName}RestClient`;
    const functionStatement = {
        isExported: true,
        name: `${clientName}`,
        parameters: [
            ...commonClientParams,
            { name: "options", type: "ClientOptions = {}" }
        ],
        returnType: clientInterfaceName,
        isDefaultExport: false,
        statements: getClientFactoryBody(clientInterfaceName, generateClientDefinition_1.pathDictionary)
    };
    if (!batch || batch.length === 1) {
        functionStatement.isDefaultExport = true;
    }
    clientFile.addFunction(functionStatement);
    clientFile.addImportDeclarations([
        {
            namedImports: ["getClient", "ClientOptions"],
            moduleSpecifier: "@azure-rest/core-client"
        }
    ]);
    clientFile.addImportDeclarations([
        {
            namedImports: credentialTypes,
            moduleSpecifier: "@azure/core-auth"
        },
        {
            namedImports: [`${clientName}RestClient`],
            moduleSpecifier: './clientDefinitions'
        }
    ]);
}
exports.generateClient = generateClient;
function getClientFactoryBody(clientTypeName, paths) {
    const { rlcShortcut } = autorestSession_1.getAutorestOptions();
    const { model } = autorestSession_1.getSession();
    const { endpoint, parameterName } = urlTransforms_1.transformBaseUrl(model);
    let baseUrl;
    if (parameterName) {
        const parsedEndpoint = endpoint === null || endpoint === void 0 ? void 0 : endpoint.replace(`{${parameterName}}`, `\${${parameterName}}`);
        baseUrl = `options.baseUrl ?? \`${parsedEndpoint}\``;
    }
    else {
        baseUrl = `options.baseUrl ?? "${endpoint}"`;
    }
    const apiVersion = getApiVersion();
    let apiVersionStatement = "";
    if (apiVersion) {
        apiVersionStatement = `options.apiVersion = options.apiVersion ?? "${apiVersion}"`;
    }
    const baseUrlStatement = {
        kind: ts_morph_1.StructureKind.VariableStatement,
        declarationKind: ts_morph_1.VariableDeclarationKind.Const,
        declarations: [{ name: "baseUrl", initializer: baseUrl }]
    };
    const { credentialScopes, credentialKeyHeaderName } = autorestSession_1.getAutorestOptions();
    const scopesString = credentialScopes && credentialScopes.length
        ? credentialScopes.map(cs => `"${cs}"`).join(", ")
        : "";
    const scopes = scopesString ? `scopes: [${scopesString}],` : "";
    const apiKeyHeaderName = credentialKeyHeaderName
        ? `apiKeyHeaderName: "${credentialKeyHeaderName}",`
        : "";
    const credentials = scopes || apiKeyHeaderName
        ? `options = {
      ...options,
      credentials: {
        ${scopes}
        ${apiKeyHeaderName}
      }
    }`
        : "";
    const getClient = `const client = getClient(
      baseUrl, ${credentials ? "credentials," : ""} options
    ) as ${clientTypeName};
    `;
    let returnStatement = `return client;`;
    if (rlcShortcut) {
        const shortcutImplementations = generateMethodShortcuts_1.generateMethodShortcutImplementation(model, paths);
        const shortcutBody = Object.keys(shortcutImplementations).map(key => {
            // If the operation group has an empty name, it means its operations are client
            // level operations so we need to spread the definitions. Otherwise they are
            // within an operation group so we add them as key: value
            return `${key ? `"${key}":` : "..."} {${shortcutImplementations[key].join()}}`;
        });
        returnStatement = `return { ...client, ${shortcutBody.join()} };`;
    }
    return [
        baseUrlStatement,
        apiVersionStatement,
        credentials,
        getClient,
        returnStatement
    ];
}
function getApiVersion() {
    const { model } = autorestSession_1.getSession();
    if (!model.globalParameters || !model.globalParameters.length) {
        return undefined;
    }
    const apiVersionParam = model.globalParameters
        .filter(gp => {
        var _a;
        return gp.implementation === codemodel_1.ImplementationLocation.Client &&
            ((_a = gp.protocol.http) === null || _a === void 0 ? void 0 : _a.in) === codemodel_1.ParameterLocation.Query;
    })
        .find(param => languageHelpers_1.getLanguageMetadata(param.language).serializedName === "api-version");
    if (apiVersionParam && schemaHelpers_1.isConstantSchema(apiVersionParam.schema)) {
        return apiVersionParam.schema.value.value.toString();
    }
    return undefined;
}
function getClientUriParameter() {
    const { model } = autorestSession_1.getSession();
    const { parameterName } = urlTransforms_1.transformBaseUrl(model);
    return parameterName;
}
//# sourceMappingURL=generateClient.js.map