"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.generateParameterInterfaces = void 0;
const codemodel_1 = require("@autorest/codemodel");
const ts_morph_1 = require("ts-morph");
const path = require("path");
const languageHelpers_1 = require("../utils/languageHelpers");
const nameUtils_1 = require("../utils/nameUtils");
const getPropertySignature_1 = require("./getPropertySignature");
const schemaHelpers_1 = require("./schemaHelpers");
const operationHelpers_1 = require("./helpers/operationHelpers");
const modelHelpers_1 = require("./helpers/modelHelpers");
const autorestSession_1 = require("../autorestSession");
/**
 * Generates the interfaces describing each operation parameters
 */
function generateParameterInterfaces(model, project) {
    var _a, _b;
    const { srcPath } = autorestSession_1.getAutorestOptions();
    const parametersFile = project.createSourceFile(path.join(srcPath, `parameters.ts`), undefined, {
        overwrite: true
    });
    // Tracks the models we need to import
    const importedModels = new Set();
    // Tracks the generated parameter types
    const operations = getAllOperations(model);
    let hasHeaders = false;
    for (const operation of operations) {
        const operationName = nameUtils_1.normalizeName(languageHelpers_1.getLanguageMetadata(operation.language).name, nameUtils_1.NameType.Interface);
        const requestCount = (_b = (_a = operation === null || operation === void 0 ? void 0 : operation.requests) === null || _a === void 0 ? void 0 : _a.length) !== null && _b !== void 0 ? _b : 0;
        const topParamName = `${operationName}Parameters`;
        const subParamNames = [];
        // We need to loop the requests. An operation with multiple requests means that
        // the operation can get different values for content-type and each value may
        // have a different type associated to it.
        for (let i = 0; i < requestCount; i++) {
            const internalReferences = new Set();
            // In case we have more than one request to model we need to add a suffix to differentiate
            const nameSuffix = i > 0 ? `${i}` : "";
            const parameterInterfaceName = requestCount > 1
                ? `${operationName}RequestParameters${nameSuffix}`
                : topParamName;
            const parameters = operationHelpers_1.getOperationParameters(operation, i);
            const queryParameterDefinitions = buildQueryParameterDefinition(operationName, parameters, [codemodel_1.SchemaContext.Input], importedModels, internalReferences, i);
            const request = operation.requests ? operation.requests[i] : undefined;
            const pathParameterDefinitions = buildPathParameterDefinitions(operationName, parameters, model, parametersFile, internalReferences, i);
            const headerParameterDefinitions = buildHeaderParameterDefinitions(operationName, parameters, parametersFile, internalReferences, i);
            const contentTypeParameterDefinition = buildContentTypeParametersDefinition(operationName, request, internalReferences, i);
            const bodyParameterDefinition = buildBodyParametersDefinition(operationName, parameters, [codemodel_1.SchemaContext.Input], importedModels, internalReferences, i);
            // Add interfaces for body and query parameters
            parametersFile.addInterfaces([
                ...(bodyParameterDefinition !== null && bodyParameterDefinition !== void 0 ? bodyParameterDefinition : []),
                ...(queryParameterDefinitions !== null && queryParameterDefinitions !== void 0 ? queryParameterDefinitions : []),
                ...(pathParameterDefinitions ? [pathParameterDefinitions] : []),
                ...(headerParameterDefinitions ? [headerParameterDefinitions] : []),
                ...(contentTypeParameterDefinition
                    ? [contentTypeParameterDefinition]
                    : [])
            ]);
            // Add Operation parameters type alias which is composed of the types we generated above
            // plus the common type RequestParameters
            parametersFile.addTypeAlias({
                name: parameterInterfaceName,
                isExported: true,
                type: [...internalReferences, "RequestParameters"].join(" & ")
            });
            subParamNames.push(parameterInterfaceName);
            if (headerParameterDefinitions !== undefined) {
                hasHeaders = true;
            }
        }
        // Add Operation parameters type alias which is composed of the types we generated above
        // plus the common type RequestParameters
        if (requestCount > 1) {
            parametersFile.addTypeAlias({
                name: topParamName,
                isExported: true,
                type: [...subParamNames].join(" | ")
            });
        }
    }
    if (hasHeaders) {
        parametersFile.addImportDeclarations([
            {
                namedImports: ["RawHttpHeadersInput"],
                moduleSpecifier: "@azure/core-rest-pipeline"
            }
        ]);
    }
    parametersFile.addImportDeclarations([
        {
            namedImports: ["RequestParameters"],
            moduleSpecifier: "@azure-rest/core-client"
        }
    ]);
    if (modelHelpers_1.hasInputModels(model)) {
        parametersFile.addImportDeclarations([
            {
                namedImports: [...importedModels],
                moduleSpecifier: "./models"
            }
        ]);
    }
}
exports.generateParameterInterfaces = generateParameterInterfaces;
function getRequestHeaderInterfaceDefinition(parameters, baseName) {
    // Check if there are any required headers
    const headerParameters = parameters.filter(p => { var _a; return ((_a = p.protocol.http) === null || _a === void 0 ? void 0 : _a.in) === "header"; });
    if (!headerParameters.length) {
        return undefined;
    }
    const headersInterfaceName = `${baseName}Headers`;
    return {
        kind: ts_morph_1.StructureKind.Interface,
        isExported: true,
        name: headersInterfaceName,
        properties: headerParameters.map((h) => {
            const description = languageHelpers_1.getLanguageMetadata(h.language).description;
            return Object.assign(Object.assign({ name: `"${languageHelpers_1.getLanguageMetadata(h.language).serializedName}"` }, (description && { docs: [{ description }] })), { type: schemaHelpers_1.primitiveSchemaToType(h.schema, [
                    codemodel_1.SchemaContext.Input,
                    codemodel_1.SchemaContext.Exception
                ]), hasQuestionToken: !h.required });
        })
    };
}
function buildHeaderParameterDefinitions(operationName, parameters, parametersFile, internalReferences, requestIndex) {
    const headerParameters = parameters.filter(p => { var _a; return ((_a = p.protocol.http) === null || _a === void 0 ? void 0 : _a.in) === "header"; });
    if (!headerParameters.length) {
        return undefined;
    }
    const nameSuffix = requestIndex > 0 ? `${requestIndex}` : "";
    const headerParameterInterfaceName = `${operationName}HeaderParam${nameSuffix}`;
    const headersInterface = getRequestHeaderInterfaceDefinition(headerParameters, operationName);
    if (headersInterface) {
        parametersFile.addInterface(headersInterface);
    }
    internalReferences.add(headerParameterInterfaceName);
    return {
        isExported: true,
        kind: ts_morph_1.StructureKind.Interface,
        name: headerParameterInterfaceName,
        properties: [
            {
                name: "headers",
                type: `RawHttpHeadersInput & ${operationName}Headers`,
                kind: ts_morph_1.StructureKind.PropertySignature
            }
        ]
    };
}
function getPathInterfaceDefinition(parameters, baseName, model) {
    // Check if there are any path parameters
    const pathParameters = parameters.filter(p => { var _a, _b; return ((_a = p.protocol.http) === null || _a === void 0 ? void 0 : _a.in) === codemodel_1.ParameterLocation.Uri && ((_b = model.globalParameters) === null || _b === void 0 ? void 0 : _b.indexOf(p)) === -1; });
    if (!pathParameters.length) {
        return undefined;
    }
    const pathInterfaceName = `${baseName}PathParameters`;
    return {
        kind: ts_morph_1.StructureKind.Interface,
        isExported: true,
        name: pathInterfaceName,
        properties: pathParameters.map((h) => {
            const description = languageHelpers_1.getLanguageMetadata(h.language).description;
            return Object.assign(Object.assign({ name: `"${languageHelpers_1.getLanguageMetadata(h.language).serializedName}"` }, (description && { docs: [{ description }] })), { type: schemaHelpers_1.primitiveSchemaToType(h.schema, [
                    codemodel_1.SchemaContext.Input,
                    codemodel_1.SchemaContext.Exception
                ]), hasQuestionToken: !h.required });
        })
    };
}
function buildPathParameterDefinitions(operationName, parameters, model, parametersFile, internalReferences, requestIndex) {
    const pathParameters = parameters.filter(p => { var _a, _b; return ((_a = p.protocol.http) === null || _a === void 0 ? void 0 : _a.in) === codemodel_1.ParameterLocation.Uri && ((_b = model.globalParameters) === null || _b === void 0 ? void 0 : _b.indexOf(p)) === -1; });
    if (!pathParameters.length) {
        return undefined;
    }
    const nameSuffix = requestIndex > 0 ? `${requestIndex}` : "";
    const pathParameterInterfaceName = `${operationName}PathParam${nameSuffix}`;
    const pathInterface = getPathInterfaceDefinition(pathParameters, operationName, model);
    if (pathInterface) {
        parametersFile.addInterface(pathInterface);
    }
    internalReferences.add(pathParameterInterfaceName);
    return {
        isExported: true,
        kind: ts_morph_1.StructureKind.Interface,
        name: pathParameterInterfaceName,
        properties: [
            {
                name: "pathParameters",
                type: `${operationName}PathParameters`,
                kind: ts_morph_1.StructureKind.PropertySignature
            }
        ]
    };
}
/**
 * Gets the interface definition for an operation bodyParameters
 */
function buildBodyParametersDefinition(operationName, parameters, schemaUsage, importedModels, internalReferences, requestIndex) {
    const bodyParameters = parameters.filter(p => { var _a; return ((_a = p.protocol.http) === null || _a === void 0 ? void 0 : _a.in) === "body"; });
    if (!bodyParameters.length) {
        return [];
    }
    const nameSuffix = requestIndex > 0 ? `${requestIndex}` : "";
    const bodyParameterInterfaceName = `${operationName}BodyParam${nameSuffix}`;
    internalReferences.add(bodyParameterInterfaceName);
    // In case of formData we'd get multiple properties in body marked as partialBody
    if (bodyParameters.some(p => p.isPartialBody)) {
        let allOptionalParts = true;
        const propertiesDefinitions = [];
        for (const param of bodyParameters) {
            if (param.required) {
                allOptionalParts = false;
            }
            propertiesDefinitions.push(getPropertySignature_1.getPropertySignature(param, schemaUsage, importedModels));
        }
        const formBodyName = `${operationName}FormBody`;
        const formBodyInterface = {
            isExported: true,
            kind: ts_morph_1.StructureKind.Interface,
            name: formBodyName,
            properties: propertiesDefinitions
        };
        return [
            {
                isExported: true,
                kind: ts_morph_1.StructureKind.Interface,
                name: bodyParameterInterfaceName,
                properties: [
                    {
                        name: "body",
                        type: formBodyName,
                        hasQuestionToken: allOptionalParts
                    }
                ]
            },
            formBodyInterface
        ];
    }
    else {
        const bodySignature = getPropertySignature_1.getPropertySignature(bodyParameters[0], schemaUsage, importedModels);
        return [
            {
                isExported: true,
                kind: ts_morph_1.StructureKind.Interface,
                name: bodyParameterInterfaceName,
                properties: [
                    {
                        docs: bodySignature.docs,
                        name: "body",
                        type: bodySignature.type,
                        hasQuestionToken: bodySignature.hasQuestionToken
                    }
                ]
            }
        ];
    }
}
/**
 * Gets the interface definition for an operation bodyParameters
 */
function buildContentTypeParametersDefinition(operationName, request, internalReferences, requestIndex) {
    var _a, _b;
    if (!request) {
        return undefined;
    }
    const mediaTypes = (_b = (_a = request.protocol.http) === null || _a === void 0 ? void 0 : _a.mediaTypes) !== null && _b !== void 0 ? _b : [];
    if (!mediaTypes.length) {
        return undefined;
    }
    const nameSuffix = requestIndex > 0 ? `${requestIndex}` : "";
    const mediaTypesParameterInterfaceName = `${operationName}MediaTypesParam${nameSuffix}`;
    // Mark the queryParameter interface for importing
    internalReferences.add(mediaTypesParameterInterfaceName);
    return {
        isExported: true,
        kind: ts_morph_1.StructureKind.Interface,
        name: mediaTypesParameterInterfaceName,
        properties: [
            {
                docs: ["Request content type"],
                name: "contentType",
                type: mediaTypes.map(mt => `"${mt}"`).join(" | "),
                hasQuestionToken: true
            }
        ]
    };
}
/**
 * Gets the interface definition for an operation queryParameters
 */
function buildQueryParameterDefinition(operationName, parameters, schemaUsage, importedModels, internalReferences, requestIndex) {
    const queryParameters = parameters.filter(p => { var _a; return ((_a = p.protocol.http) === null || _a === void 0 ? void 0 : _a.in) === "query"; });
    if (!queryParameters.length) {
        return undefined;
    }
    const nameSuffix = requestIndex > 0 ? `${requestIndex}` : "";
    const queryParameterInterfaceName = `${operationName}QueryParam${nameSuffix}`;
    const queryParameterPropertiesName = `${operationName}QueryParamProperties`;
    // Get the property signature for each query parameter
    const propertiesDefinition = queryParameters.map(qp => getPropertySignature_1.getPropertySignature(qp, schemaUsage, importedModels));
    const hasRequiredParameters = propertiesDefinition.some(p => !p.hasQuestionToken);
    const propertiesInterface = {
        kind: ts_morph_1.StructureKind.Interface,
        isExported: true,
        name: queryParameterPropertiesName,
        properties: propertiesDefinition
    };
    const parameterInterface = {
        kind: ts_morph_1.StructureKind.Interface,
        isExported: true,
        name: queryParameterInterfaceName,
        properties: [
            {
                name: "queryParameters",
                type: queryParameterPropertiesName,
                // Mark as optional if there are no required parameters
                hasQuestionToken: !hasRequiredParameters
            }
        ]
    };
    // Mark the queryParameter interface for importing
    internalReferences.add(queryParameterInterfaceName);
    return [propertiesInterface, parameterInterface];
}
/**
 * Flattens all operations from operationGroups
 */
function getAllOperations(model) {
    const operations = [];
    for (const operationGroup of model.operationGroups) {
        operations.push(...operationGroup.operations);
    }
    return operations;
}
//# sourceMappingURL=generateParameterTypes.js.map