"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.getFormatDocs = exports.isConstantSchema = exports.isDictionarySchema = exports.primitiveSchemaToType = exports.getElementType = void 0;
const codemodel_1 = require("@autorest/codemodel");
const languageHelpers_1 = require("../utils/languageHelpers");
const nameUtils_1 = require("../utils/nameUtils");
/**
 * Get the Typescript type to generate from an Schema
 * @param schema - Schema to find element type for
 * @param importedModels - Set to track models to import
 */
function getElementType(schema, schemaUsage, importedModels = new Set()) {
    if (isArraySchema(schema)) {
        // Recursively find out the type for the elements in the array.
        return `Array<${getElementType(schema.elementType, schemaUsage, importedModels)}>`;
    }
    if (isPrimitiveSchema(schema)) {
        return `${primitiveSchemaToType(schema, schemaUsage)}`;
    }
    if (isAnyObjectSchema(schema)) {
        return `Record<string, unknown>`;
    }
    if (codemodel_1.isObjectSchema(schema)) {
        const nameSuffix = schemaUsage.includes(codemodel_1.SchemaContext.Output)
            ? "Output"
            : "";
        const schemaInfo = getObjectInfo(schema);
        const name = `${schemaInfo.name}${nameSuffix}`;
        importedModels.add(name);
        return `${name}`;
    }
    if (isDictionarySchema(schema)) {
        return `Record<string, ${getElementType(schema.elementType, schemaUsage, importedModels)}>`;
    }
    throw new Error(`Don't know how to get type for schema ${schema.type}`);
}
exports.getElementType = getElementType;
/**
 * Determine whether or not a schema maps to a Typescript primitive type.
 */
function isPrimitiveSchema(schema) {
    return [
        codemodel_1.SchemaType.Binary,
        codemodel_1.SchemaType.String,
        codemodel_1.SchemaType.Number,
        codemodel_1.SchemaType.Integer,
        codemodel_1.SchemaType.Date,
        codemodel_1.SchemaType.DateTime,
        codemodel_1.SchemaType.Any,
        codemodel_1.SchemaType.Boolean,
        codemodel_1.SchemaType.ByteArray,
        codemodel_1.SchemaType.Char,
        codemodel_1.SchemaType.Credential,
        codemodel_1.SchemaType.Duration,
        codemodel_1.SchemaType.Time,
        codemodel_1.SchemaType.UnixTime,
        codemodel_1.SchemaType.Uri,
        codemodel_1.SchemaType.Uuid,
        codemodel_1.SchemaType.Unknown,
        codemodel_1.SchemaType.Constant,
        codemodel_1.SchemaType.Choice,
        codemodel_1.SchemaType.SealedChoice
    ].includes(schema.type);
}
/**
 * Given a primitive schema, determines the Typescript type to generate
 * @param schema - Schema to generate a type for
 * @returns a string with the Typescript type to generate for the given schema
 */
function primitiveSchemaToType(schema, schemaUse) {
    switch (schema.type) {
        case codemodel_1.SchemaType.Any:
            return "any";
        case codemodel_1.SchemaType.Integer:
        case codemodel_1.SchemaType.Number:
            return "number";
        case codemodel_1.SchemaType.Date:
        case codemodel_1.SchemaType.Time:
        case codemodel_1.SchemaType.DateTime:
            return schemaUse.includes(codemodel_1.SchemaContext.Output)
                ? "string"
                : "Date | string";
        case codemodel_1.SchemaType.Char:
            return "string";
        case codemodel_1.SchemaType.ByteArray:
        case codemodel_1.SchemaType.Duration:
        case codemodel_1.SchemaType.Credential:
        case codemodel_1.SchemaType.UnixTime:
        case codemodel_1.SchemaType.Uri:
        case codemodel_1.SchemaType.Uuid:
        case codemodel_1.SchemaType.String:
            return "string";
        case codemodel_1.SchemaType.Binary:
            return schemaUse.includes(codemodel_1.SchemaContext.Output)
                ? "Uint8Array"
                : "string | Uint8Array";
        case codemodel_1.SchemaType.Boolean:
            return "boolean";
        case codemodel_1.SchemaType.Choice:
        case codemodel_1.SchemaType.SealedChoice:
            return schema.choices
                .map(choice => `"${choice.value}"`)
                .join(" | ");
        case codemodel_1.SchemaType.Constant:
            const value = schema.value.value;
            return typeof value === "string" ? `"${value}"` : value;
    }
    throw new Error(`Unknown primitive schema ${schema.type}`);
}
exports.primitiveSchemaToType = primitiveSchemaToType;
/**
 * Gets object schema metadata
 * @param schema - Schema definition of an object
 */
function getObjectInfo(schema) {
    const name = nameUtils_1.normalizeName(languageHelpers_1.getLanguageMetadata(schema.language).name, nameUtils_1.NameType.Interface, true);
    return {
        name
    };
}
function isArraySchema(schema) {
    return schema.type === codemodel_1.SchemaType.Array;
}
function isAnyObjectSchema(schema) {
    return schema.type === codemodel_1.SchemaType.AnyObject;
}
function isDictionarySchema(schema) {
    return schema.type === codemodel_1.SchemaType.Dictionary;
}
exports.isDictionarySchema = isDictionarySchema;
function isConstantSchema(schema) {
    return schema.type === codemodel_1.SchemaType.Constant;
}
exports.isConstantSchema = isConstantSchema;
/**
 * Gets additional documentation for a given schema
 */
function getFormatDocs(schema) {
    switch (schema.type) {
        case codemodel_1.SchemaType.ByteArray:
            return "Value may contain base64 encoded characters";
        case codemodel_1.SchemaType.Binary:
            return "Value may contain any sequence of octets";
        case codemodel_1.SchemaType.Credential:
            return "Value may contain a password";
        case codemodel_1.SchemaType.Uuid:
            return "Value may contain a UUID";
        case codemodel_1.SchemaType.Uri:
            return "Value may contain a URL";
        default:
            return undefined;
    }
}
exports.getFormatDocs = getFormatDocs;
//# sourceMappingURL=schemaHelpers.js.map