"use strict";
// Copyright (c) Microsoft Corporation.
// Licensed under the MIT License.
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.generateTypeScriptLibrary = void 0;
const prettier = require("prettier");
const fsextra = require("fs-extra");
const path = require("path");
const ts_morph_1 = require("ts-morph");
const transforms_1 = require("./transforms/transforms");
const samplesTransforms_1 = require("./transforms/samplesTransforms");
const clientFileGenerator_1 = require("./generators/clientFileGenerator");
const modelsGenerator_1 = require("./generators/modelsGenerator");
const mappersGenerator_1 = require("./generators/mappersGenerator");
const indexGenerator_1 = require("./generators/indexGenerator");
const packageFileGenerator_1 = require("./generators/static/packageFileGenerator");
const apiExtractorConfig_1 = require("./generators/static/apiExtractorConfig");
const licenseFileGenerator_1 = require("./generators/static/licenseFileGenerator");
const readmeFileGenerator_1 = require("./generators/static/readmeFileGenerator");
const sampleTestGenerator_1 = require("./generators/test/sampleTestGenerator");
const tsConfigFileGenerator_1 = require("./generators/static/tsConfigFileGenerator");
const rollupConfigFileGenerator_1 = require("./generators/static/rollupConfigFileGenerator");
const operationGenerator_1 = require("./generators/operationGenerator");
const operationInterfaceGenerator_1 = require("./generators/operationInterfaceGenerator");
const sampleEnvGenerator_1 = require("./generators/samples/sampleEnvGenerator");
const sampleGenerator_1 = require("./generators/samples/sampleGenerator");
const parametersGenerator_1 = require("./generators/parametersGenerator");
const LROGenerator_1 = require("./generators/LROGenerator");
const tracingFileGenerator_1 = require("./generators/tracingFileGenerator");
const autorestSession_1 = require("./autorestSession");
const conflictResolver_1 = require("./conflictResolver");
const prettierTypeScriptOptions = {
    parser: "typescript",
    arrowParens: "always",
    bracketSpacing: true,
    endOfLine: "lf",
    printWidth: 80,
    semi: true,
    singleQuote: false,
    tabWidth: 2
};
const prettierJSONOptions = {
    parser: "json",
    tabWidth: 2,
    semi: false,
    singleQuote: false
};
function generateTypeScriptLibrary(codeModel, host) {
    var _a, _b;
    return __awaiter(this, void 0, void 0, function* () {
        const project = new ts_morph_1.Project({
            useInMemoryFileSystem: true,
            manipulationSettings: {
                indentationText: ts_morph_1.IndentationText.TwoSpaces
            }
        });
        const { packageDetails, licenseHeader: shouldGenerateLicense, generateTest, generateSample, outputPath, srcPath } = autorestSession_1.getAutorestOptions();
        const clientDetails = yield transforms_1.transformCodeModel(codeModel);
        conflictResolver_1.conflictResolver(clientDetails);
        if (generateSample) {
            clientDetails.samples = yield samplesTransforms_1.transformSamples(codeModel, clientDetails);
        }
        // Skip metadata generation if `generate-metadata` is explicitly false
        packageFileGenerator_1.generatePackageJson(project, clientDetails);
        licenseFileGenerator_1.generateLicenseFile(project);
        readmeFileGenerator_1.generateReadmeFile(clientDetails, project);
        if (generateTest) {
            sampleTestGenerator_1.generateSampleTestFile(project);
        }
        tsConfigFileGenerator_1.generateTsConfig(project);
        rollupConfigFileGenerator_1.generateRollupConfig(project);
        apiExtractorConfig_1.generateApiExtractorConfig(project);
        clientFileGenerator_1.generateClient(clientDetails, project);
        modelsGenerator_1.generateModels(clientDetails, project);
        mappersGenerator_1.generateMappers(clientDetails, project);
        operationGenerator_1.generateOperations(clientDetails, project);
        operationInterfaceGenerator_1.generateOperationsInterfaces(clientDetails, project);
        if (generateSample && ((_a = clientDetails === null || clientDetails === void 0 ? void 0 : clientDetails.samples) === null || _a === void 0 ? void 0 : _a.length) && ((_b = clientDetails === null || clientDetails === void 0 ? void 0 : clientDetails.samples) === null || _b === void 0 ? void 0 : _b.length) > 0) {
            sampleGenerator_1.generateSamples(clientDetails, project);
            sampleEnvGenerator_1.generateSampleEnv(project);
        }
        parametersGenerator_1.generateParameters(clientDetails, project);
        indexGenerator_1.generateIndexFile(project, clientDetails);
        yield LROGenerator_1.generateLroFiles(clientDetails, project);
        tracingFileGenerator_1.generateTracingFile(project);
        const licenseHeader = `
/*
 * Copyright (c) Microsoft Corporation.
 * Licensed under the MIT License.
 *
 * Code generated by Microsoft (R) AutoRest Code Generator.
 * Changes may cause incorrect behavior and will be lost if the code is regenerated.
 */
`;
        // Save the source files to the virtual filesystem
        project.saveSync();
        const fs = project.getFileSystem();
        const pathToClear = outputPath ? path.join(outputPath, srcPath) : srcPath;
        fsextra.emptyDirSync(`${pathToClear}`);
        // Loop over the files
        for (const file of project.getSourceFiles()) {
            const filePath = file.getFilePath();
            const isJson = /\.json$/gi.test(filePath);
            const isSourceCode = /\.(ts|js)$/gi.test(filePath);
            let fileContents = fs.readFileSync(filePath);
            // Add the license header to source code files
            if (shouldGenerateLicense && isSourceCode) {
                fileContents = `${licenseHeader.trimLeft()}\n${fileContents}`;
            }
            // Format the contents if necessary
            if (isJson || isSourceCode) {
                fileContents = prettier.format(fileContents, isJson ? prettierJSONOptions : prettierTypeScriptOptions);
            }
            // Write the file to the AutoRest host
            host.writeFile({
                filename: filePath.substr(1),
                content: fileContents
            });
        }
    });
}
exports.generateTypeScriptLibrary = generateTypeScriptLibrary;
//# sourceMappingURL=typescriptGenerator.js.map